#ifndef UTILS_LOGGER_SPDLOG_DETAILS_WINDOWS_INCLUDE_H_
#define UTILS_LOGGER_SPDLOG_DETAILS_WINDOWS_INCLUDE_H_

#ifndef NOMINMAX
#define NOMINMAX // prevent windows redefining min/max
#endif

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>
#endif  // UTILS_LOGGER_SPDLOG_DETAILS_WINDOWS_INCLUDE_H_
